#! /bin/ksh

# Initialize global variables
kernel=""
image=""
string_match=""
outfile=""
sysstats=0
tracedata=0
verbose=1
ans="y"
DEBUGIO=0
set -A PacketPoolTags
PacketPoolSize=0
set -A ExecStatsTags
ExecStatsSize=0

##############################################################################
# SCRIPT NAME: kthreads.ksh
##############################################################################
#
# DESCRIPTION:
#
# This script is used to gather data concerning the state of the DFS
# services on a machine, or from a system image.  It uses crash to get 
# some of the kernel data structures that have DFS performance data (if
# this has been requested), as well as gather the stacks of the various 
# threads that are performing DFS services.
# 
##############################################################################


##############################################################################
# FUNCTION: usage
##############################################################################
#
# This function prints the proper usage of the script.
#
##############################################################################
usage() {
  print "ERROR:Unknown option/bad syntax:"
  print "usage: kthreads [-k kernel] [-i system_image] [-q] [-s] [-t]"
  print "\t\t-o output_file "
  print "\t -k: specify the kernel image to use, default /unix"
  print "\t -i: specify the system image to use, default current"
  print "\t -q: quiet"
  print "\t -s: gather system stats, specify the level of DCE:"
  print "\t\t   "
  print "\t -t: gather tracing output"
  exit 1
}
# END: usage
##############################################################################

##############################################################################
# FUNCTION: ProcessArgs
##############################################################################
#
# This function processes the command line args, displaying any relevant
# error messages.
#
##############################################################################
ProcessArgs () {

  while getopts ":k:i:o:qst" opt; do
    case $opt in
      k )
        if [[ $verbose -ne 0 ]]; then
          print "Using kernel image $OPTARG."
        fi
        kernel=$OPTARG ;;
      i )  
        if [[ $verbose -ne 0 ]]; then
          print "Using system image $OPTARG."
        fi
        image=$OPTARG ;;
      o )  
        if [[ $verbose -ne 0 ]]; then
          print "Using output file $OPTARG."
        fi
        outfile=$OPTARG ;;
      q )
        #
        # print "Setting verbose output to 1."
        # I guess if we want it quiet, why let 'em know?
        #
        verbose=0 ;;
      s )
        if [[ $verbose -ne 0 ]]; then
          print "Going to gather system stats (netstat, iostat, vmstat)."
        fi
        sysstats=1 ;;
      t )
        if [[ $verbose -ne 0 ]]; then
          print "Going to gather DFS tracing output."
        fi
        tracedata=1 ;;
      \? ) usage ;;
    esac
  done
  shift $(($OPTIND - 1))

  #
  # Must specify the string to match against the threads in the system image.
  #
  if [[ -z $outfile ]]; then
    print "ERROR: Must specify an output file!:"
    usage
  fi

  #
  # If kernel image not specified.
  #
  if [[ -z $kernel && $verbose -ne 0 ]]; then
    print "Using the current system kernel."
  fi

  # 
  # Set the system image to nothing
  #
  if [[ -z $image && $verbose -ne 0 ]]; then
    print "Using the current system image."
  fi

  set -A filters $@
}
# END: ProcessArgs
##############################################################################

##############################################################################
# FUNCTION: CleanUp
##############################################################################
#
# This function cleans up any temporary files & ??
#
##############################################################################
CleanUp () {
  rm -rf /tmp/$$
  if [[ $? -ne 0 ]]; then
    print "Unable to remove temporary files!"
  fi
}
# END: CleanUp
##############################################################################

##############################################################################
# FUNCTION: InitPacketPoolArray
##############################################################################
#
# This function initializes the array used for printing out packet pools stats
#
##############################################################################
InitPacketPoolArray () {
  PacketPoolTags[0]="pkt_mutex              "
  PacketPoolTags[1]="version                "
  PacketPoolTags[2]="active_rqe             "
  PacketPoolTags[3]="active_xqe             "
  PacketPoolTags[4]="reservations           "
  PacketPoolTags[5]="mem_allocated          "
  PacketPoolTags[6]="mem_ration_level       "
  PacketPoolTags[7]="mem_peak_level         "
  PacketPoolTags[8]="max_reservations       "
  PacketPoolTags[9]="srv_reservations       "
  PacketPoolTags[10]="srv_resv_avail         "
  PacketPoolTags[11]="noqueue_resv           "
  PacketPoolTags[12]="active_ccall_resv      "
  PacketPoolTags[13]="active_scall_resv      "
  PacketPoolTags[14]="active_blocked_resv    "
  PacketPoolTags[15]="total_reservations     "
  PacketPoolTags[16]="total_sonly_resv       "
  PacketPoolTags[17]="total_queued_resv      "
  PacketPoolTags[18]="total_blocked_resv     "
  PacketPoolTags[19]="total_noavail_resv     "
  PacketPoolTags[20]="failed_alloc_rqe       "
  PacketPoolTags[21]="blocked_alloc_rqe      "
  PacketPoolTags[22]="rqe_overflows          "
  PacketPoolTags[23]="xqe_mbuf_refs          "
  PacketPoolTags[24]="active_rqes_peak       "
  PacketPoolTags[25]="active_xqes_peak       "
  PacketPoolTags[26]="xqe_throttles          "
  PacketPoolTags[27]="mbuf_member_offset     "
  PacketPoolTags[28]="fack_pkt_size          "
  PacketPoolTags[29]="fack_pkt_size_increases"
  PacketPoolTags[30]="fack_pkt_size_decreases"
  PacketPoolTags[31]="large_pkt_size         "
  PacketPoolTags[32]="larg_pkt_window_size   "
  PacketPoolTags[33]="large_pkt_blast_size   " 
  PacketPoolTags[34]="low_pkts_window_size   "
  PacketPoolTags[35]="pkt_gc_count           "
  PacketPoolTags[36]="pkt_gc_timer           "
  PacketPoolTags[37]="gc_pkts                "
  PacketPoolTags[38]="gc_bytes               "
  PacketPoolTags[39]="max_window             "
  PacketPoolTags[40]="medium_window          "
  PacketPoolTags[41]="small_window           "
  PacketPoolTags[42]="low_pkts               "
  PacketPoolTags[43]="rationing              "
  PacketPoolSize=44
}

##############################################################################
# FUNCTION: DumpPacketPool
##############################################################################
#
# This function gathers information from the packet pool data structure.
#
##############################################################################
DumpPacketPool () {
  InitPacketPoolArray
  typeset -L2 PoolPre
  print -p "!print \
    \"========================================================\" >> $outfile "
  print -p "!print \
    \"=========       Packet Pool Statistics         =========\" >> $outfile "
  print -p "!print \
    \"========================================================\" >> $outfile "
  print -p "!print \"\" >> $outfile "
  print -p "knlist rpc_g_dg_pkt_pool"
  read -p PoolHeader PoolAddr
  PoolPre=${PoolAddr}
  while [[ $PoolPre != "0x" ]]; do
    if [[ $PoolAddr = "no match" ]]; then
      print -p \
        "!print \"No match for rpc_g_dg_pkt_pool in dump\" >> $outfile "
      return
    fi
    if [[ $DEBUGIO -eq 1 ]]; then
      print "Bad data $PoolAddr, skipping"
    fi
    read -p PoolHeader PoolAddr
    PoolPre=${PoolAddr}
  done
  if [[ $DEBUGIO -eq 1 ]]; then
    print "PoolAddr $PoolAddr"  
  fi
  let counter=0
  let loops=${PacketPoolSize}/4
  print -p "od ${PoolAddr} ${PacketPoolSize} D"
  while [[ $counter -lt $loops ]]; do
    read -p addr one two three four
    let ind1=4*$counter
    let ind2=4*$counter+1
    let ind3=4*$counter+2
    let ind4=4*$counter+3
#    print -p "!print \" ${PacketPoolTags[ind1]} = $one\" >> $outfile"
#    print -p "!print \" ${PacketPoolTags[ind2]} = $two\" >> $outfile"
    if [[ ${#one} -ge 6 ]]; then
      print -p "!print \" ${PacketPoolTags[ind1]} = $one\
      ${PacketPoolTags[ind2]} = $two \" >> $outfile"
    elif [[ ${#one} -ge 3 ]]; then
      print -p "!print \" ${PacketPoolTags[ind1]} = $one\
      \t${PacketPoolTags[ind2]} = $two \" >> $outfile"
    else
      print -p "!print \" ${PacketPoolTags[ind1]} = $one\
      \t\t${PacketPoolTags[ind2]} = $two \" >> $outfile"
    fi
#    print -p "!print \" ${PacketPoolTags[ind3]} = $three \" >> $outfile"
#    print -p "!print \" ${PacketPoolTags[ind4]} = $four\" >> $outfile"
    if [[ ${#three} -ge 6 ]]; then
      print -p "!print \" ${PacketPoolTags[ind3]} = $three\
      ${PacketPoolTags[ind4]} = $four\" >> $outfile"
    elif [[ ${#three} -ge 3 ]]; then
      print -p "!print \" ${PacketPoolTags[ind3]} = $three\
      \t${PacketPoolTags[ind4]} = $four\" >> $outfile"
    else
      print -p "!print \" ${PacketPoolTags[ind3]} = $three\
      \t\t${PacketPoolTags[ind4]} = $four\" >> $outfile"
    fi
    let counter=$counter+1
  done
}

##############################################################################
# FUNCTION: InitExecStatsArray
##############################################################################
#
# This function initializes the array used for printing out execution stats
#
##############################################################################
InitExecStatsArray () {
  ExecStatsTags[0]="exec_calls          "
  ExecStatsTags[1]="stub_calls          "
  ExecStatsTags[2]="gt1sec_calls        "
  ExecStatsTags[3]="gt10sec_calls       "
  ExecStatsTags[4]="gt60sec_calls       "
  ExecStatsTags[5]="longest_call        "
  ExecStatsTags[6]="longest_opnum       "
  ExecStatsTags[7]="longest_call_tstamp "
  ExecStatsTags[8]="waycalls            "
  ExecStatsTags[9]="wayerrs             "
  ExecStatsTags[10]="gt10sec_waycalls    "
  ExecStatsTags[11]="gt60sec_waycalls    "
  ExecStatsTags[12]="gt120sec_waycalls   "
  ExecStatsTags[13]="waycalls_queued     "
  ExecStatsTags[14]=""
  ExecStatsTags[15]=""
  ExecStatsSize=16
}

##############################################################################
# FUNCTION: DumpExecStats
##############################################################################
#
# This function gathers information from the execution statistics structure.
#
##############################################################################
DumpExecStats () {
  InitExecStatsArray
  typeset -L2 ExecPre
  print -p "!print \"\" >> $outfile "
  print -p "!print \"\" >> $outfile "
  print -p "!print \
    \"========================================================\" >> $outfile "
  print -p "!print \
    \"=========         Execution Statistics         =========\" >> $outfile "
  print -p "!print \
    \"========================================================\" >> $outfile "
  print -p "!print \"\" >> $outfile "
  print -p "knlist rpc_g_dg_exec_stats"
  read -p ExecHeader ExecStatsAddr
  ExecPre=${ExecStatsAddr}
  while [[ $ExecPre != "0x" ]]; do
    if [[ $DEBUGIO -eq 1 ]]; then
      print "Bad data $ExecStatsAddr, skipping"
    fi
    if [[ $ExecStatsAddr = "no match" ]]; then
      print -p \
        "!print \"No match for rpc_g_dg_exec_stats in dump\" >> $outfile "
      return
    fi
    read -p ExecHeader ExecStatsAddr
    ExecPre=${ExecStatsAddr}
  done
  if [[ $DEBUGIO -eq 1 ]]; then
    print "ExecStatsAddr $ExecStatsAddr"  
  fi
  let counter=0
  let loops=${ExecStatsSize}/4
  print -p "od ${ExecStatsAddr} ${ExecStatsSize} D"
  while [[ $counter -lt $loops ]]; do
    read -p addr one two three four
    let ind1=4*$counter
    let ind2=4*$counter+1
    let ind3=4*$counter+2
    let ind4=4*$counter+3
    if [[ ${#one} -ge 6 ]]; then
      print -p "!print \" ${ExecStatsTags[ind1]} = $one\
      ${ExecStatsTags[ind2]} = $two \" >> $outfile"
    elif [[ ${#one} -ge 3 ]]; then
      print -p "!print \" ${ExecStatsTags[ind1]} = $one\
      \t${ExecStatsTags[ind2]} = $two \" >> $outfile"
    else
      print -p "!print \" ${ExecStatsTags[ind1]} = $one\
      \t\t${ExecStatsTags[ind2]} = $two \" >> $outfile"
    fi
    if [[ ${#ExecStatsTags[ind3]} -ge 1 ]]; then
      if [[ ${#three} -ge 6 ]]; then
        print -p "!print \" ${ExecStatsTags[ind3]} = $three\
      ${ExecStatsTags[ind4]} = $four\" >> $outfile"
      elif [[ ${#three} -ge 3 ]]; then
        print -p "!print \" ${ExecStatsTags[ind3]} = $three\
      \t${ExecStatsTags[ind4]} = $four\" >> $outfile"
      else
        print -p "!print \" ${ExecStatsTags[ind3]} = $three\
      \t\t${ExecStatsTags[ind4]} = $four\" >> $outfile"
      fi
    fi
    let counter=$counter+1
  done
}

##############################################################################
# FUNCTION: AllThreads
##############################################################################
#
# This function gathers information for all threads in the system image.
#
##############################################################################
AllThreads () {

  print -p "!print \"\" >> $outfile "
  print -p "!print \
    \"========================================================\" >> $outfile "
  print -p "!print \
    \"=====  START: Thread information for system image  =====\" >> $outfile "
  print -p "!print \"\" >> $outfile "
  print -p "thread | egrep ^[\ 1-9][\ 0-9][0-9] | cut -c1-$MAXSIZE"
  print -p "calc A"
  while read -p thread_number
  do
    #-------------------------------------------------------------------
    # Check for our token "a" that is eof for us
    #-------------------------------------------------------------------
    if [[ $thread_number = a ]]; then
      if [[ $DEBUGIO -eq 1 ]]; then
        print "Found token exiting loop"
      fi
      break
    fi
  
    #-------------------------------------------------------------------
    # This should filter out all stdio data that is not a thread number
    #-------------------------------------------------------------------
    if [[ ${#thread_number} -gt $MAXSIZE ]]; then
      if [[ $DEBUGIO -eq 1 ]]; then
        print "We have a problem with $thread_number - skip..."
      fi
      continue
    fi
  
    #-------------------------------------------------------------------
    # Otherwise, dump out the thread information.
    #-------------------------------------------------------------------
    if [[ $DEBUGIO -eq 1 ]]; then
      print "Getting info for thread ${thread_number}"
    fi
    print -p "!print \"\" >> $outfile"
    print -p \
      "!print \"***** Thread info, slot ${thread_number} *****\" >> $outfile"
    print -p "thread ${thread_number} >> $outfile"
    print -p "t -k ${thread_number} >> $outfile"
  done
  
  print -p "!print \"\" >> $outfile"
  print -p "!print \"=====    END: Thread information for system image    =====\" >> $outfile"
  print -p "!print \"==========================================================\" >> $outfile"
  if [[ $DEBUGIO -eq 1 ]]; then
    print "Escaped from the loop"
  fi
}
# END: AllThreads
##############################################################################

##############################################################################
# FUNCTION: FilterThreads
##############################################################################
#
# This function gathers information for all threads in the system image.
#
##############################################################################
FilterThreads () {

  print -p "!print \"\" >> $outfile "
  print -p "!print \
    \"========================================================\" >> $outfile "
  print -p "!print \
    \"=====  START: Thread information for system image  =====\" >> $outfile "
  print -p "!print \"\" >> $outfile "

  let index=0
  for filter in ${filters[*]}
  do
    print -p "!print \"=== Threads matching filter $filter.===\" >> $outfile"
    print -p "thread | egrep ^[\ 1-9][\ 0-9][0-9] | egrep $filter | cut -c1-$MAXSIZE"
    print -p "calc A"
    while read -p thread_number
    do
      #-------------------------------------------------------------------
      # Check for our token "a" that is eof for us
      #-------------------------------------------------------------------
      if [[ $thread_number = a ]]; then
        if [[ $DEBUGIO -eq 1 ]]; then
          print "Found token exiting loop"
        fi
        break
      fi
  
      #-------------------------------------------------------------------
      # This should filter out all stdio data that is not a thread number
      #-------------------------------------------------------------------
      if [[ ${#thread_number} -gt $MAXSIZE ]]; then
        if [[ $DEBUGIO -eq 1 ]]; then
          print "We have a problem with $thread_number - skip..."
        fi
        continue
      fi
  
      #-------------------------------------------------------------------
      # Otherwise, dump out the thread information.
      #-------------------------------------------------------------------
      if [[ $DEBUGIO -eq 1 ]]; then
        print "Getting info for thread ${thread_number}"
      fi
      print -p "!print \"\" >> $outfile"
      print -p \
        "!print \"***** Thread info, slot ${thread_number} *****\" >> $outfile"
      print -p "thread ${thread_number} >> $outfile"
      print -p "t -k ${thread_number} >> $outfile"
    done
    let index=index+1
  done 

  print -p "!print \"\" >> $outfile"
  print -p "!print \"=====    END: Thread information for system image    =====\" >> $outfile"
  print -p "!print \"==========================================================\" >> $outfile"

  if [[ $DEBUGIO -eq 1 ]]; then
    print "Escaped from the loop"
  fi
}
# END: FilterThreads
##############################################################################
  
##############################################################################
# Main Script
##############################################################################

#
# Maximum size of string expected fro thread ID
#
MAXSIZE=4

ProcessArgs $@
if [[ $? -ne 0 ]]; then
  exit 1
fi

#
# Add date to output file name to make sure it is unique
#
#outfile=`date +$outfile.%m.%d.%y_%H.%M`

if [[ -f $outfile ]]; then
  print -n "Remove current $outfile? (y/n)"
  read ans
  if [[ "$ans" = "y" ]]; then
    rm $outfile
    if [[ $? -ne 0 ]]; then
      print "Unable to remove $outfile!\n"
      exit 1
    fi
  else
    print "Aborting..."
    exit 1
  fi
else
  touch $outfile
  if [[ $? -ne 0 ]]; then
    print "Unable to create $outfile!\n"
    exit 1
  fi
fi

if [[ $tracedata -eq 1 ]]; then
  dfstrace clear -log cmfx
  dfstrace setset -set krpc -active
fi

#
# Open up pipe with crash
#
crash $image $kernel 2>/dev/null|&

#
# Want to get the DFS statistics using internal data structures
# Get the address with knlist, strip off the "{symbol name}:" that
# is also output by knlist
#

if [[ $sysstats -ne 0 ]]; then
  if [[ $verbose -ne 0 ]]; then
    print "Getting data from kernel data structures..."
  fi
  DumpPacketPool
  DumpExecStats
  if [[ $verbose -ne 0 ]]; then
    print "Gathered data from kernel data structures..."
  fi
fi

if [[ ${#filters[*]} -gt 0 ]]; then
   print "Gathering filtered data about from system image."
#  for filter in ${filters[*]}
#  do
    FilterThreads $filter
#  done
else
  AllThreads
fi
print -p "quit"
print "Done"